/* 
 *  psplash - a lightweight framebuffer splashscreen for embedded devices. 
 *
 *  Copyright (c) 2006 Matthew Allum <mallum@o-hand.com>
 *  Copyright (c) 2010 Florian Boor <florian@kernelconcepts.de>
 *  Copyright (c) 2012 Daniel Wagener <daniel@kernelconcepts.de>
 *
 *  Parts of this file ( fifo handling ) based on 'usplash' copyright 
 *  Matthew Garret.
 *  PPM file parsing from Busybox,  
 *  Copyright (C) 2008 Michele Sanges <michele.sanges@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */

#include <stdint.h>
#include <sys/types.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>

#include "psplash.h"
#include "psplash-fb.h"
#include "psplash-image.h"
#include "radeon-font.h"
#include "rect2987-img.h"
#include "psplash-console.h"
#include "psplash-poky-img.h"

#define MSG ""
#define BUFFER_SIZE 4096

static int    img_width;
static int    img_height;
static int    img_bytes_per_pixel;
static uint8* img_pixel_data;
static int    psplash_bg_r = PSPLASH_BACKGROUND_COLOR_R;
static int    psplash_bg_g = PSPLASH_BACKGROUND_COLOR_G;
static int    psplash_bg_b = PSPLASH_BACKGROUND_COLOR_B;
static int    psplash_progress_r = PSPLASH_PROGRESS_COLOR_R;
static int    psplash_progress_g = PSPLASH_PROGRESS_COLOR_G;
static int    psplash_progress_b = PSPLASH_PROGRESS_COLOR_B;
static int    psplash_txt_r = PSPLASH_TEXT_COLOR_R;
static int    psplash_txt_g = PSPLASH_TEXT_COLOR_G;
static int    psplash_txt_b = PSPLASH_TEXT_COLOR_B;
static image *bar_bg;
static int    bar_x;
static int    bar_y;
static int    bar_width;
static int    bar_height;
static int    disable_bar;

const char *
skip_whitespace (const char *s)
{
  while (isspace(*s)) 
    {
      ++s;
    }

  return s;
}


static void
psplash_read_ppm (const char *name)
{
    unsigned int  width, height, line_size, i, j;
    int           hicolor = 0;
    char          *ptr, *head, *pixline;
    char          buffer[BUFFER_SIZE] = {0};
    unsigned char *p;
    FILE          *theme_file;

    theme_file = fopen (name, "r");
    if (theme_file == NULL)
      {
        perror("Unable to open:");
        return;
      }

    head = malloc(256);

  /* parse ppm header
   * - A ppm image’s magic number is the two characters "P6".
   * - Whitespace (blanks, TABs, CRs, LFs).
   * - A width, formatted as ASCII characters in decimal.
   * - Whitespace.
   * - A height, again in ASCII decimal.
   * - Whitespace.
   * - The maximum color value (Maxval), again in ASCII decimal. Must be
   *   less than 65536.
   * - Newline or other single whitespace character.
   * - A raster of Width * Height pixels in triplets of rgb
   *   in pure binary by 1 (or not implemented 2) bytes.
   */
  while (1) 
    {
      if (fgets(head, 255, theme_file) == NULL
          /* do not overrun the buffer */
          || strlen(buffer) >= sizeof(buffer) - 256)
        {
          perror("uh?");
          fprintf(stderr, "Bad PPM file header '%s'\n", name);
          return;
        }

    ptr = memchr(skip_whitespace(head), '#', 256);
    if (ptr != NULL)
      {
        *ptr = 0; /* ignore comments */
      }
    strcat(buffer, head);
    /* width, height, max_color_val */
    if (sscanf(buffer, "P6 %u %u %u", &width, &height, &i) == 3)
      {
        if (i > 255)
          {
            hicolor = 1;
          }
        break;
      }
    /* If we do not find a signature throughout the whole file then
       we will diagnose this via EOF on read in the head of the loop.  */
    }

  free(head);

  if (hicolor) 
    {
      fprintf (stderr, "High color PPM unimplemented\n");
      return ;
#if 0    
      line_size = width * 6;
      img_pixel_data = malloc(width * height * 3);
      p = img_pixel_data;
      pixline = malloc(line_size);
      for (j = 0; j < height; j++)
        {
          char *pixel = pixline;

          if (fread(pixline, 1, line_size, theme_file) != line_size)
            {
              fprintf(stderr, "Bad PPM file '%s'\n", name);
            }
          for (i = 0; i < width; i++)
            {
              unsigned thispix;
              thispix = (((unsigned)pixel[0] << 8) & 0xf800)
              | (((unsigned)pixel[1] << 3) & 0x07e0)
              | (((unsigned)pixel[2] >> 3));
              *p++ = thispix;
              pixel += 3;
            }
        }
#endif
    }
  else
    {
      line_size = width * 3;
      img_pixel_data = malloc(line_size * height);
      p = img_pixel_data;
      pixline = malloc(line_size);
      for (j = 0; j < height; j++) 
        {
          if (fread(pixline, 1, line_size, theme_file) != line_size)
            {
              fprintf(stderr, "Bad PPM file '%s'\n", name);
            }
          for (i = 0; i < width * 3; i++)
            {
              *p++ = pixline[i];
            }
        }
    }

  free(pixline);
  fclose(theme_file);
  img_height = height;
  img_width = width;
  img_bytes_per_pixel = 3;
}

void
psplash_exit ()
{
  DBG("mark", );

  psplash_console_reset ();
}

void
psplash_draw_msg (PSplashFB *fb, const char *msg)
{
  int w, h;

  psplash_fb_text_size (fb, &w, &h, &radeon_font, msg);

  DBG("displaying '%s' %ix%i\n", msg, w, h);

  /* Clear */

  psplash_fb_draw_rect (fb, 0, 
                        fb->height - (fb->height * PSPLASH_TEXT_DEFAULT / 100) - h,
                        fb->width, h, psplash_bg_r, psplash_bg_g, psplash_bg_b);

  psplash_fb_draw_text (fb, (fb->width-w)/2,
                        fb->height - (fb->height * PSPLASH_TEXT_DEFAULT / 100) - h,
                        psplash_txt_r, psplash_txt_g, psplash_txt_b,
                        &radeon_font,
                        msg);
}

void
psplash_draw_progress (PSplashFB *fb, int value)
{
  int barwidth;

  if (value > 0)
    {
      barwidth = (CLAMP(value,0,100) * bar_width) / 100;
    }
  else
    {
      barwidth = (CLAMP(-value,0,100) * bar_width) / 100;
    }

  psplash_fb_draw_status_bar (fb, bar_x, bar_y, bar_width, bar_height,
                     psplash_progress_r, psplash_progress_g, psplash_progress_b,
                                                              barwidth, bar_bg);


  DBG("value: %i, width: %i, barwidth :%i\n", value, bar_width, barwidth);
}

static int
parse_command (PSplashFB *fb, char *string, int length)
{
  char *command;
  int   parsed=0;

  parsed = strlen(string)+1;

  DBG("got cmd %s", string);

  if (strcmp(string,"QUIT") == 0)
    {
      return 1;
    }

  command = strtok(string," ");

  if (!strcmp(command,"PROGRESS"))
    {
      if (!disable_bar)
          psplash_draw_progress (fb, atoi(strtok(NULL,"\0")));
    }
  else if (!strcmp(command,"MSG"))
    {
      psplash_draw_msg (fb, strtok(NULL,"\0"));
    }
  else if (!strcmp(command,"QUIT"))
    {
      return 1;
    }

  return 0;
}

void
psplash_main (PSplashFB *fb, int pipe_fd, int timeout)
{
  int            err;
  ssize_t        length = 0;
  fd_set         descriptors;
  struct timeval tv;
  char          *end;
  char           command[2048];

  tv.tv_sec = timeout;
  tv.tv_usec = 0;

  FD_ZERO(&descriptors);
  FD_SET(pipe_fd, &descriptors);

  end = command;

  while (1)
    {
      if (timeout != 0)
        {
          err = select(pipe_fd+1, &descriptors, NULL, NULL, &tv);
        }
      else
        {
          err = select(pipe_fd+1, &descriptors, NULL, NULL, NULL);
        }

      if (err <= 0)
        {
          /*
          if (errno == EINTR)
            continue;
          */
          return;
        }

      length += read (pipe_fd, end, sizeof(command) - (end - command));

      if (length == 0) 
        {
          /* Reopen to see if there's anything more for us */
          close(pipe_fd);
          pipe_fd = open(PSPLASH_FIFO,O_RDONLY|O_NONBLOCK);
          goto out;
        }

      if (command[length-1] == '\0')
        {
          if (parse_command(fb, command, strlen(command)))
            {
              return;
            }
          length = 0;
        }
      else if (command[length-1] == '\n')
        {
          command[length-1] = '\0';
          if (parse_command(fb, command, strlen(command)))
            {
              return;
            }
          length = 0;
        }


    out:
      end = &command[length];

      tv.tv_sec = timeout;
      tv.tv_usec = 0;

      FD_ZERO(&descriptors);
      FD_SET(pipe_fd,&descriptors);
    }

  return;
}

int 
main (int argc, char** argv) 
{
  char      *tmpdir;
  char      *filename = NULL;
  int        pipe_fd, i = 0, angle = 0, ret = 0;
  PSplashFB *fb;
  bool      disable_console_switch = FALSE;
  int       logo_xrel = PSPLASH_LOGO_X_DEFAULT;
  int       logo_yrel = PSPLASH_LOGO_Y_DEFAULT;
  int       bar_xrel  = PSPLASH_PROGRESS_X_DEFAULT;
//  int       text_xrel = PSPLASH_TEXT_X_DEFAULT;
  int       bar_yrel  = PSPLASH_PROGRESS_Y_DEFAULT;
//  int       text_yrel = PSPLASH_TEXT_Y_DEFAULT;
        
  signal(SIGHUP, psplash_exit);
  signal(SIGINT, psplash_exit);
  signal(SIGQUIT, psplash_exit);

  while (++i < argc)
    {
      if (!strcmp(argv[i],"-n") || !strcmp(argv[i],"--no-console-switch"))
        {
          disable_console_switch = TRUE;
          continue;
        }
      if (!strcmp(argv[i],"-d") || !strcmp(argv[i],"--disable-bar"))
        {
          disable_bar = TRUE;
          continue;
        }
      if (!strcmp(argv[i],"-a") || !strcmp(argv[i],"--angle"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          angle = atoi(argv[i]);
          continue;
        }
      if (!strcmp(argv[i],"-b") || !strcmp(argv[i],"--background"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          filename = argv[i];
          continue;
        }
      if (!strcmp(argv[i],"-c") || !strcmp(argv[i],"--bgcolor"))
        {
          int r, g, b;
          if (++i > argc)
            {
              goto fail;
            }
          if (!argv[i])
            {
              goto fail;
            }
          if (sscanf(argv[i],"%02x%02x%02x", &r, &g, &b) != 3)
            {
              goto fail;
            }
          psplash_bg_r = r;
          psplash_bg_g = g;
          psplash_bg_b = b;
          continue;
        }
      if (!strcmp(argv[i],"-p") || !strcmp(argv[i],"--progresscolor"))
        {
          int r, g, b;
          if (++i > argc)
            {
              goto fail;
            }
          if (!argv[i])
            {
              goto fail;
            }
          if (sscanf(argv[i],"%02x%02x%02x", &r, &g, &b) != 3)
            {
              goto fail;
            }
          psplash_progress_r = r;
          psplash_progress_g = g;
          psplash_progress_b = b;
          continue;
        }
      if (!strcmp(argv[i],"-t") || !strcmp(argv[i],"--textcolor"))
        {
          int r, g, b;
          if (++i > argc)
            {
              goto fail;
            }
          if (!argv[i])
            {
              goto fail;
            }
          if (sscanf(argv[i],"%02x%02x%02x", &r, &g, &b) != 3)
            {
              goto fail;
            }
          psplash_txt_r = r;
          psplash_txt_g = g;
          psplash_txt_b = b;
          continue;
        }
      if (!strcmp(argv[i],"--logo-x"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          logo_xrel = atoi(argv[i]);
          continue;
        }
      if (!strcmp(argv[i],"--logo-y"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          logo_yrel = atoi(argv[i]);
          continue;
        }
      if (!strcmp(argv[i],"--bar-x"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          bar_xrel = atoi(argv[i]);
          continue;
        }
      if (!strcmp(argv[i],"--bar-y"))
        {
          if (++i > argc)
            {
              goto fail;
            }
          bar_yrel = atoi(argv[i]);
          continue;
        }

    fail:
      fprintf(stderr,
        "Usage: %s [-n|--no-console-switch][-a|--angle <0|90|180|270>] [-b|--background <filename.ppm>]\n" \
           "[-c|--bgcolor rrggbb] [-p|--progresscolor rrggbb] [-t|--textcolor rrggbb] [-d|--disable-bar]\n" \
           "[--logo-x \%\%][--logo-y \%\%][--bar-x \%\%][--bar-y \%\%]\n\n",
      argv[0]);
      exit(-1);
    }

  tmpdir = getenv("TMPDIR");

  if (!tmpdir)
    {
      tmpdir = "/tmp";
    }

  if (chdir(tmpdir))
    {
      perror("tmpdir");
      exit(-1);
    }

  if (mkfifo(PSPLASH_FIFO, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP))
    {
      if (errno!=EEXIST)
        {
          perror("mkfifo");
          exit(-1);
        }
    }

  pipe_fd = open (PSPLASH_FIFO,O_RDONLY|O_NONBLOCK);

  if (pipe_fd==-1)
    {
      perror("pipe open");
      exit(-2);
    }

  if (!disable_console_switch)
    {
      psplash_console_switch ();
    }

  if ((fb = psplash_fb_new(angle)) == NULL)
    {
      ret = -1;
      goto fb_fail;
    }

  /* Read ppm file if passed to us */
  if (filename) 
    {
      psplash_read_ppm (filename);
    }

  /* Clear the background if necessary */
  if ((fb->width != img_width) || (fb->height != img_height))
    {
      psplash_fb_draw_rect (fb, 0, 0, fb->width, fb->height, psplash_bg_r,
                            psplash_bg_g, psplash_bg_b);
    }

  /* Draw the logo  */
  if (img_height && img_width)
    {
      psplash_fb_draw_image_simple (fb, (fb->width  - img_width) * logo_xrel / 100,
                                    (fb->height - img_height) * logo_yrel / 100,
                                    img_width, img_height, img_bytes_per_pixel,
                                    img_pixel_data);

    } 
  else 
    {
      psplash_fb_draw_image_rle (fb,
                                 (fb->width  - POKY_IMG_WIDTH) * logo_xrel / 100,
                                 (fb->height - POKY_IMG_HEIGHT) * logo_yrel / 100,
                                 POKY_IMG_WIDTH, POKY_IMG_HEIGHT,
                                 POKY_IMG_BYTES_PER_PIXEL,
                                 POKY_IMG_RLE_PIXEL_DATA);
    }

  /* 4 pix border */
  bar_x      = ((fb->width - BAR_IMG_WIDTH) * bar_xrel / 100) + 4 ;
  bar_y      = fb->height - (fb->height * bar_yrel / 100) + 4;
  bar_width  = BAR_IMG_WIDTH - 8;
  bar_height = BAR_IMG_HEIGHT - 8;

  /* Draw progress bar border */
  if (!disable_bar) 
    {
      psplash_fb_draw_image_rle_round_corners (fb,
                             (fb->width  - BAR_IMG_WIDTH) * bar_xrel / 100,
                             fb->height - (fb->height * bar_yrel / 100),
                             BAR_IMG_WIDTH, BAR_IMG_HEIGHT,
                             BAR_IMG_BYTES_PER_PIXEL, BAR_IMG_RLE_PIXEL_DATA);
      bar_bg = image_new_from_fb (fb, bar_x, bar_y, bar_width, bar_height);

      psplash_draw_progress (fb, 0);
    }

  if (MSG && strlen(MSG))
    {
      psplash_draw_msg (fb, MSG);
    }

  psplash_main (fb, pipe_fd, 0);

  image_destroy (bar_bg);

  psplash_fb_destroy (fb);

fb_fail:
  unlink(PSPLASH_FIFO);

  if (!disable_console_switch)
    {
      psplash_console_reset ();
    }

  return ret;
}

